'use strict';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

// Returns the width, height and offsets of the provided element
var getElementClientRect = (element => {
  const {
    width,
    height,
    top,
    left
  } = element.getBoundingClientRect();
  return {
    width,
    height,
    y: top,
    x: left
  };
});

var getParentNode = (element => {
  if (element.nodeName === 'HTML') {
    // DocumentElement detectedF
    return element;
  }

  return element.parentNode || // DOM Element detected
  // $FlowFixMe: need a better way to handle this...
  element.host || // ShadowRoot detected
  document.ownerDocument || // Fallback to ownerDocument if available
  document.documentElement // Or to documentElement if everything else fails
  ;
});

function getScrollParent(node) {
  if (!node) {
    return document.body;
  }

  switch (node.nodeName) {
    case 'HTML':
    case 'BODY':
      return node.ownerDocument.body;

    case '#document':
      // Flow doesn't understand nodeName type refinement unfortunately
      return node.body;
  }

  if (node instanceof Element) {
    // Firefox want us to check `-x` and `-y` variations as well
    const {
      overflow,
      overflowX,
      overflowY
    } = getComputedStyle(node);

    if (/(auto|scroll|overlay)/.test(overflow + overflowY + overflowX)) {
      return node;
    }
  }

  return getScrollParent(getParentNode(node));
}

function listScrollParents(element, list = []) {
  const scrollParent = getScrollParent(element);
  const isBody = scrollParent.nodeName === 'BODY';
  const target = isBody ? scrollParent.ownerDocument.defaultView : scrollParent;
  const updatedList = list.concat(target);
  return isBody ? updatedList : updatedList.concat(listScrollParents(getParentNode(target)));
}

function getWindow(element) {
  const ownerDocument = element.ownerDocument;
  return ownerDocument ? ownerDocument.defaultView : window;
}

var unwrapJqueryElement = (element => element && element.jquery ? element[0] : element);

const order = modifiers => {
  let sortedModifiers = [...modifiers];
  let i = 0,
      j,
      temp;

  while (i < sortedModifiers.length) {
    temp = sortedModifiers.slice(0, i);

    for (j = i; j < sortedModifiers.length; j++) {
      if ((sortedModifiers[j].requires || []).every(n => temp.some(({
        name
      }) => n === name))) {
        sortedModifiers.splice(i++, 0, sortedModifiers.splice(j, 1)[0]);
        break;
      }
    }
  }

  return sortedModifiers;
};

var orderModifiers = (modifiers => [...order(modifiers.filter(({
  phase
}) => phase === 'read')), ...order(modifiers.filter(({
  phase
}) => phase === 'main')), ...order(modifiers.filter(({
  phase
}) => phase === 'write'))]);

// Expands the eventListeners value to an object containing the
// `scroll` and `resize` booleans
//
// true => true, true
// false => false, false
// true, false => true, false
// false, false => false, false
var expandEventListeners = (eventListeners => {
  const fallbackValue = typeof eventListeners === 'boolean' ? eventListeners : false;
  return {
    scroll: typeof eventListeners.scroll === 'boolean' ? eventListeners.scroll : fallbackValue,
    resize: typeof eventListeners.resize === 'boolean' ? eventListeners.resize : fallbackValue
  };
});

var getBasePlacement = (placement => placement.split('-')[0]);

const top = 'top';
const bottom = 'bottom';
const right = 'right';
const left = 'left';
const basePlacements = [top, bottom, right, left];
const placements = basePlacements.reduce((acc, placement) => acc.concat([placement, `${placement}-start`, `${placement}-end`]), []); // modifiers that need to read the DOM

const read = 'read'; // pure-logic modifiers

const main = 'main'; // modifier with the purpose to write to the DOM (or write into a framework state)

const write = 'write';

var computeOffsets = (({
  reference,
  popper,
  strategy,
  placement
}) => {
  const basePlacement = getBasePlacement(placement);

  switch (basePlacement) {
    case top:
      return {
        x: reference.x + reference.width / 2 - popper.width / 2,
        y: reference.y - popper.height
      };

    case bottom:
      return {
        x: reference.x + reference.width / 2 - popper.width / 2,
        y: reference.y + reference.height
      };

    case right:
      return {
        x: reference.x + reference.width,
        y: reference.y + reference.height / 2 - popper.height / 2
      };

    case left:
      return {
        x: reference.x - popper.width,
        y: reference.y + reference.height / 2 - popper.height / 2
      };

    default:
      // $FlowFixMe: This will actually never match; github.com/facebook/flow/issues/2395
      return undefined;
  }
});

var format = ((str, ...args) => [...args].reduce((p, c) => p.replace(/%s/, c), str));

function microtaskDebounce(fn) {
  let called = false;
  return () => new Promise(resolve => {
    if (called) {
      return resolve();
    }

    called = true;
    Promise.resolve().then(() => {
      called = false;
      resolve(fn());
    });
  });
}

const ERROR_MESSAGE = 'PopperJS: modifier "%s" provided an invalid %s property, expected %s but got %s';
const VALID_PROPERTIES = ['name', 'enabled', 'phase', 'fn', 'onLoad', 'requires', 'options'];
var validateModifiers = (modifiers => {
  modifiers.forEach(modifier => {
    Object.keys(modifier).forEach(key => {
      switch (key) {
        case 'name':
          if (typeof modifier.name !== 'string') {
            console.error(format(ERROR_MESSAGE, String(modifier.name), '"name"', '"string"', `"${String(modifier.name)}"`));
          }

          break;

        case 'enabled':
          if (typeof modifier.enabled !== 'boolean') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"enabled"', '"boolean"', `"${String(modifier.enabled)}"`));
          }

        case 'phase':
          if (![read, main, write].includes(modifier.phase)) {
            console.error(format(ERROR_MESSAGE, modifier.name, '"phase"', 'either "read", "main" or "write"', `"${String(modifier.phase)}"`));
          }

          break;

        case 'fn':
          if (typeof modifier.fn !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"fn"', '"function"', `"${String(modifier.fn)}"`));
          }

          break;

        case 'onLoad':
          if (typeof modifier.onLoad !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"onLoad"', '"function"', `"${String(modifier.fn)}"`));
          }

          break;

        case 'requires':
          if (!Array.isArray(modifier.requires)) {
            console.error(format(ERROR_MESSAGE, modifier.name, '"requires"', '"array"', `"${String(modifier.requires)}"`));
          }

          break;

        case 'object':
          break;

        default:
          console.error(`PopperJS: an invalid property has been provided to the "${modifier.name}" modifier, valid properties are ${VALID_PROPERTIES.map(s => `"${s}"`).join(', ')}; but "${key}" was provided.`);
      }
    });
  });
});

const areValidElements = (a, b) => a instanceof Element && b instanceof Element;

const defaultOptions = {
  placement: 'bottom',
  eventListeners: {
    scroll: true,
    resize: true
  },
  modifiers: []
};
class Popper {
  constructor(reference, popper, options = defaultOptions) {
    _defineProperty(this, "state", {
      reference: undefined,
      popper: undefined,
      orderedModifiers: [],
      measures: {},
      offsets: {},
      scrollParents: {},
      options: defaultOptions
    });

    // make update() debounced, so that it only runs at most once-per-tick
    this.update = microtaskDebounce(this.update.bind(this)); // Unwrap `reference` and `popper` elements in case they are
    // wrapped by jQuery, otherwise consume them as is

    this.state.reference = unwrapJqueryElement(reference);
    this.state.popper = unwrapJqueryElement(popper);
    const {
      reference: referenceElement,
      popper: popperElement
    } = this.state; // Store options into state

    this.state.options = _objectSpread({}, defaultOptions, options); // Don't proceed if `reference` or `popper` are invalid elements

    if (!areValidElements(referenceElement, popperElement)) {
      return;
    }

    this.state.scrollParents = {
      reference: listScrollParents(referenceElement),
      popper: listScrollParents(popperElement)
    }; // Order `options.modifiers` so that the dependencies are fulfilled
    // once the modifiers are executed

    this.state.orderedModifiers = orderModifiers(this.state.options.modifiers); // Validate the provided modifiers so that the consumer will get warned
    // of one of the custom modifiers is invalid for any reason

    {
      validateModifiers(this.state.options.modifiers);
    } // Modifiers have the opportunity to execute some arbitrary code before
    // the first update cycle is ran, the order of execution will be the same
    // defined by the modifier dependencies directive.
    // The `onLoad` function may add or alter the options of themselves


    this.state.orderedModifiers.forEach(({
      onLoad,
      enabled
    }) => enabled && onLoad && onLoad(this.state));
    this.update().then(() => {
      // After the first update completed, enable the event listeners
      this.enableEventListeners(this.state.options.eventListeners);
    });
  } // Async and optimistically optimized update
  // it will not be executed if not necessary
  // check Popper#constructor to see how it gets debounced


  update() {
    return new Promise((success, reject) => {
      this.forceUpdate();
      success(this.state);
    });
  } // Syncronous and forcefully executed update
  // it will always be executed even if not necessary, usually NOT needed
  // use Popper#update instead


  forceUpdate() {
    const {
      reference: referenceElement,
      popper: popperElement
    } = this.state; // Don't proceed if `reference` or `popper` are not valid elements anymore

    if (!areValidElements(referenceElement, popperElement)) {
      return;
    } // Get initial measurements
    // these are going to be used to compute the initial popper offsets
    // and as cache for any modifier that needs them later


    this.state.measures = {
      reference: getElementClientRect(referenceElement),
      popper: getElementClientRect(popperElement)
    };
    this.state.offsets = {
      popper: computeOffsets({
        reference: this.state.measures.reference,
        popper: this.state.measures.popper,
        strategy: 'absolute',
        placement: this.state.options.placement
      })
    };
    this.state.orderedModifiers.forEach(({
      fn,
      enabled
    }) => enabled && fn && (this.state = fn(this.state)));
  }

  enableEventListeners(eventListeners) {
    const {
      reference: referenceElement,
      popper: popperElement
    } = this.state;
    const {
      scroll,
      resize
    } = expandEventListeners(eventListeners); // Don't proceed if `reference` or `popper` are not valid elements anymore

    if (!areValidElements(referenceElement, popperElement)) {
      return;
    }

    if (scroll) {
      const scrollParents = [...listScrollParents(referenceElement), ...listScrollParents(popperElement)];
      scrollParents.length > 0 && scrollParents.forEach(scrollParent => scrollParent.addEventListener('scroll', this.update, {
        passive: true
      }));
    }

    if (resize && this.state.popper) {
      const win = getWindow(this.state.popper);
      win && win.addEventListener('resize', this.update, {
        passive: true
      });
    }
  }

}

module.exports = Popper;
//# sourceMappingURL=index.js.map
